/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rtp.h"
#include "in.h"
#include "sdp.h"

struct insession_descr {
    struct sdp_descr sdp;
    char *url;
    int active;
    struct insession_descr *next;
    struct insession *in;

    struct rtp_session *out[MAX_MEDIA];
};

static struct insession_descr *first_active_insession;

static int session_handle_input(int fd, struct rtp_session **h, int rtcp)
{
    struct rtp_session *r;
    static uint8_t buff[4096];
    int res;

    res = read(fd, buff, 4096);
    if (res < 0) {
        perror("Read");

        return -1;
    }
    if (res > 1500) {
        fprintf(stderr, "Uhmmm... %d\n", res);
    }
    if (res == 4096) {
        return -1;
    }
    r = *h;

    while (r) {
        if (rtcp) {
            rtcp_send(r, buff, res);
        } else {
            /* RTP */
            rtp_send(r, buff, res);
        }
        r = r->next;
    }

    return 0;
}

static int session_handle_rtp_input(int fd, void *h)
{
    return session_handle_input(fd, h, 0);
}

static int session_handle_rtcp_input(int fd, void *h)
{
    return session_handle_input(fd, h, 1);
}

/*
 * Insession Methods
 */
static struct insession *insession_priv_get(const char *url)
{
    struct insession_descr *s;

    s = first_active_insession;
    while (s != NULL) {
        if (strcmp(s->url, url) == 0) {
            s->active++;

            return s->in;
	}

        s = s->next;
    }

    return NULL;
}

static struct insession *insession_priv_allocate(const char *url)
{
    struct insession *s;
    
    s = malloc(sizeof(struct insession) + sizeof(struct insession_descr));
    memset(s, 0, sizeof(struct insession) + sizeof(struct insession_descr));
    s->ops = &IS;
    s->ctx = (struct insession_descr *)((uint8_t *)s + sizeof(struct insession));
    s->ctx->next = s->ctx;
    s->ctx->url = strdup(url);
    s->ctx->in = s;

    return s;
}

static void release(struct insession_descr *s)
{
    struct insession_descr *p;
    int i;

    if (s->active == 0) {
        for (i = 0; i < MAX_MEDIA; i++) {
            if (s->sdp.media[i]) {
                rtp_insession_release(s->sdp.media[i]->in);
                /*rtp_session_close(s->sdp.media[i]->in); already done... */
            }
        }
        sdp_free(&s->sdp);
        if (s == first_active_insession) {
            first_active_insession = s->next;
        }
        p = first_active_insession;
        while (p != NULL) {
            if (p->next == s) {
                p->next = s->next;
	        p = NULL;		/* exit from while */
            } else {
                p = p->next;
            }
        }

        free(s->url);
	free(s->in);

	return;
    }

    s->active--;
}

static void activate(struct insession_descr *s, struct event_handler *h)
{
    if (s->next == s) {
        int i;
        if (s->active != 0) {
            fprintf(stderr, "Error: activating out-of-list session with act=%d\n", s->active);
            abort();
        }
        s->next = first_active_insession;
        first_active_insession = s;
	/* TODO: Open the input sockets!!! */
	i = 0;
        while (s->sdp.media[i]) {
            s->sdp.media[i]->in = rtp_insession_create(h, *s->sdp.media[i]->dst_addr ? s->sdp.media[i]->dst_addr : s->sdp.dst_addr, s->sdp.media[i]->port);
            if (s->sdp.media[i]->in == NULL) {
                fprintf(stderr, "Cannot open input session!\n");
            } else {
                rtp_insession_handle(s->sdp.media[i]->in, session_handle_rtp_input, session_handle_rtcp_input, &s->out[i]);
            }
            i++;
	}
    }
    /* s->active++ is not needed here, because it has already been done in session_get() */
}

static const char *get_url(struct insession_descr *s)
{
    return s->url;
}

static int get_sdp(char *buff, int size, struct insession_descr *s)
{
    return sdp_get(buff, size, &s->sdp);
}

static int set_sdp(struct insession_descr *s, const char *sdp)
{
    return sdp_set(&s->sdp, sdp);
}

static struct media *media(struct insession_descr *s, int i)
{
    if ((i > MAX_MEDIA) || (s == NULL)) {
        return NULL;
    }

    return s->sdp.media[i];
}

static void play(struct insession_descr *s, struct rtp_session *r, int i)
{
    r->next = s->out[i];
    s->out[i] = r;
}

static void stop(struct insession_descr *s, struct rtp_session *r)
{
    int i;

    for (i = 0; i < MAX_MEDIA; i++) {
        if (s->out[i] == r) {
            s->out[i] = r->next;
            return;
        } else {
            struct rtp_session *p;

            p = s->out[i];
            while(p != NULL) {
                if (p->next == r) {
                    p->next = r->next;
                    return;
                } else {
                    p = p->next;
                }
            }
        }
    }
}

struct insession_methods IS = {
    .get = insession_priv_get,
    .allocate = insession_priv_allocate,
    .release = release,
    .activate = activate,
    .get_url = get_url,
    .get_sdp = get_sdp,
    .set_sdp = set_sdp,
    .media = media,
    .play = play,
    .stop = stop,
};
