/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "rtp.h"
#include "net.h"
#include "engine.h"
#include "parse.h"
#include "rtsp_commands.h"
#include "stats.h"
#include "dbg.h"

#define BSIZE 1024

struct rtsp_connection {
    struct event_handler *context;
    char *buffer;
    char *start;
    char *end;
    int from;
};

static int tag_cmp(const char *buf, const char *tag_name)
{
    return (strncmp(buf, tag_name, strlen(tag_name)) == 0);
}

static const char *request_complete_text(const char *buff)
{
    char *p;

    p = strstr(buff, "\n\n");
    if (p != NULL) {
        return p + 2;
    }
    p = strstr(buff, "\r\n\r\n");
    if (p != NULL) {
        return p + 4;
    }

    return NULL;
}

static const char *request_complete_rtp(const char *buff, const char *end)
{
    if ((buff[0] == '$') && (end - buff > 4)) {
        int len;

        len = buff[2];
        len = len << 8;
        len |= buff[3];

        if (len + 4 <= end - buff) {
            return buff + len + 4;
        }
    }

    return NULL;
}

/* Fill stream_name and track_id by parsing the URL */
/* FIXME: This is so ugly that it is not even fun... Should be rewritten! */
static void rtsp_url_parse(char *abs_path, const char *url)
{
    const char *p;

    if (!tag_cmp(url, "rtsp:/") && !tag_cmp(url, "RTSP:/")) {
        return;
    }
    p = url + 6;
    if (p[0] == '/') {
        /* Skip the <server>[:port] part... */
        p += 1;
	while ((*p != '/') && (*p != 0)) {
            p++;
	}
    }

    if (*p == 0) {
      strcpy(abs_path, "/");

      return;
    }

    strncpy(abs_path, p + 1, RTSP_PARAM_STRING_MAX - 1);
}

static const char *rtsp_parse_first_line(const char *buff, struct rtsp_request *r)
{
    const char *p;
    char cmd_name[RTSP_PARAM_STRING_MAX];
    char url[RTSP_PARAM_STRING_MAX];
    char protocol[RTSP_PARAM_STRING_MAX];

    p = word_get(cmd_name, buff);
    p = word_get(url, p);
    p = word_get(protocol, p);

    if (!tag_cmp(protocol, "RTSP/1.0")) {
        return NULL;
    }
    r->command = rtsp_command(cmd_name);
#if 0
    if (r->command == -1) {
        return NULL;
    }
#endif
    rtsp_url_parse(r->abs_path, url);

    dbg_print("U:| %s | %s\n", r->abs_path, url);
    return p;
}

static void rtsp_parse_transport(const char *buff, struct rtsp_request *r)
{
    while (*buff && (*buff != '\r') && (*buff != '\n')) {
        buff = spaces_skip(buff);
        if (tag_cmp(buff, "RTP/AVP/TCP")) {
            r->streaming_mode = RTP_TCP;
        } else if ((tag_cmp(buff, "RTP/AVP")) ||
            (tag_cmp(buff, "RTP/AVP/UDP"))) {
            r->streaming_mode = RTP_UDP;
        } else if (tag_cmp(buff, "RAW/RAW/UDP")) {
            r->streaming_mode = RAW_UDP;
        } else if (tag_cmp(buff, "MP2T/H2221/UDP")) {
            r->streaming_mode = TS_UDP;
        } else if (tag_cmp(buff, "destination=")) {
            r->dst = strdup(buff + 12);
        } else if (tag_cmp(buff, "ttl")) {
            r->ttl = atoi(buff + 4);
        } else if (tag_cmp(buff, "client_port=")) {
            int res;

            res = sscanf(buff + 12, "%hu-%hu", &r->rtp_port, &r->rtcp_port);
            if (res != 2) {
                res = sscanf(buff + 12, "%hu", &r->rtp_port);
                if (res == 1) {
                    if ((r->streaming_mode == RAW_UDP) || (r->streaming_mode == TS_UDP)) {
                        r->rtcp_port = 0;
                    } else {
                        r->rtcp_port = r->rtp_port + 1;
                    }
                } else {
                    r->rtcp_port = 0;
                    r->rtp_port = 0;
                }
            }
        } else if (tag_cmp(buff, "interleaved=")) {
            int res;

            res = sscanf(buff + 12, "%u-%u", &r->rtp_channel_id, &r->rtcp_channel_id);
            if (res != 2) {
                r->rtp_channel_id = 0;
                r->rtcp_channel_id = 0;
            }
        } else if (tag_cmp(buff, "multicast")) {
            r->streaming_mode |= MULTICAST;
	}

        while ((*buff != 0) && (*buff != ';')) {
            buff++;
        }
	if (*buff) {
            buff++;
	}
    }
}

static void rtsp_parse_range_npt(const char *buff, struct rtsp_request *r)
{
}
static const char *rtsp_parse_line(const char *buff, struct rtsp_request *r)
{
    const char *p;

    p = buff;
    if (tag_cmp(p, "Session:")) {
        p += 8;
	p = spaces_skip(p);
        r->session_id = atol(p);
    } else if (tag_cmp(p, "Content-Length:")) {
        p += 15;
	p = spaces_skip(p);
        r->content_length = atol(p);
    } else if (tag_cmp(p, "Transport:")) {
        p += 10;
	p = spaces_skip(p);
        rtsp_parse_transport(p, r);

	/* TODO: Fix this hack! */
	if (r->streaming_mode & 9) {
            fprintf(stderr, "Hack: Disable MULTICAST!\n");
            r->streaming_mode = RTP_UDP;
        }
    } else if (tag_cmp(p, "CSeq:")) {
        p += 5;
	p = spaces_skip(p);
        r->cseq = atol(p);
    } else if (tag_cmp(p, "Range:")) {
        p += 6;
	p = spaces_skip(p);
        rtsp_parse_range_npt(p, r);
    }

    return p;
}

static struct rtsp_request *request_parse(const char *buff, int len)
{
    static struct rtsp_request req;
    const char *p;

    free(req.dst);
    memset(&req, 0, sizeof(struct rtsp_request));
    dbg_print("Received %d: %s\n", len, buff);
    p = rtsp_parse_first_line(buff, &req);
    if (p == NULL) {
        return NULL;
    }
    p = line_skip(p);
    while(p) {
        p = rtsp_parse_line(p, &req);
        p = line_skip(p);
    }

    return &req;
}

static int rtsp_handle_connection(int fd, void *c)
{
    int len;
    struct rtsp_connection *connection;
    const char *p;

    connection = c;
    if (connection->buffer == NULL) {
        connection->buffer = malloc(BSIZE + 1);
        connection->start = connection->buffer;
    }
    if (connection->buffer == NULL) {
        return -1;
    }
    
    len = recv(fd, connection->start, BSIZE - (connection->start - connection->buffer), 0);
    if (len < 0) {
        return -1;
    }

    if (len == 0) {
        /* Other end of the socket is closed? */
        close(fd);
        event_del(connection->context, fd);
        free(connection->buffer);
        free(connection);

	return 0;
    }
    connection->start[len] = 0;
    connection->start += len;
    p = request_complete_text(connection->buffer);
    if (p) {
        static char response[RTSP_PARAM_STRING_MAX];	/* FIXME */
	struct rtsp_request *r;

        r = request_parse(connection->buffer, p - connection->buffer);
	if (r != NULL) {
	    r->handle = connection->context;
            r->fd = fd;
            switch (r->command) {
                case RTSP_CMD_OPTIONS:
                    rtsp_cmd_options(response, RTSP_PARAM_STRING_MAX, r);
                    break;
                case RTSP_CMD_DESCRIBE:
                    rtsp_cmd_describe(response, RTSP_PARAM_STRING_MAX, r);
                    break;
                case RTSP_CMD_SETUP:
                    r->from = connection->from;
                    r->local_addr = fd_addr(fd);
                    rtsp_cmd_setup(response, RTSP_PARAM_STRING_MAX, r);
                    /*connection->session_id = r->session_id;*/
                    break;
                case RTSP_CMD_PLAY:
                    rtsp_cmd_play(response, RTSP_PARAM_STRING_MAX, r);
		    break;
                case RTSP_CMD_TEARDOWN:
                    rtsp_cmd_teardown(response, RTSP_PARAM_STRING_MAX, r);
		    break;
                default:
                    fprintf(stderr, "Unknown command %d:\n%s\n",
                            r->command, connection->buffer);
                    rtsp_cmd_unknown(response, RTSP_PARAM_STRING_MAX, r);
            }
        } else {
            rtsp_cmd_error(response, RTSP_PARAM_STRING_MAX);
        }
        dbg_print("Sending %s\n", response);
	send(fd, response, strlen(response), 0);
    } else {
        p = request_complete_rtp(connection->buffer, connection->start);
    }

    if (p) {
        memmove(connection->buffer, p, connection->start - p);
        connection->start -= (p - connection->buffer);
    }

    return 0;
}

static int rtsp_new_connection(int s, void *c)
{
    struct rtsp_connection *connection;
    int res, fd, client_ip;

    fd = new_connection(s, &client_ip);
    connection = malloc(sizeof(struct rtsp_connection));
    if (connection == NULL) {
        perror("MAlloc");
        close(fd);

        return -1;
    }
    memset(connection, 0, sizeof(struct rtsp_connection));
    connection->context = c;
    connection->from = client_ip;
    res = event_new(c, fd, rtsp_handle_connection, connection);
    if (res < 0) {
        fprintf(stderr, "Cannot create connection event!");
        close(fd);
	free(connection);

        return -1;
    }

    return 0;
}

int server_create(const char *iface, int port)
{
    int s, res, done;
    struct event_handler *h;

    s = server_init(iface, port);
    if (s < 0) {
        return s;
    }
    h = event_handler_init();
    if (h == NULL) {
        fprintf(stderr, "Cannot create event handler\n");
        close(s);

        return -1;
    }
    res = event_new(h, s, rtsp_new_connection, h);
    if (res < 0) {
        fprintf(stderr, "Cannot create server event!");
        event_handler_cleanup(h);
        close(s);

        return -1;
    }
    tlog_init(h, 8080);		/* This creates a telnet server for logging */

    done = 0;
    while(!done) {
        event_loop(h);
    }

    return 0;
}
