#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "sdp.h"
#include "parse.h"

struct attribute {
    char *key;
    char *val;
};

static char *line_copy(const char *p)
{
    const char *p1;
    char *res;

    p1 = line_skip(p);
    if (p1 == NULL) {
        return NULL;
    }
    p1 -= 1;
    res = malloc(p1 - p + 1);
    if (res == NULL) {
        return NULL;
    }
    memcpy(res, p, (p1 - p));
    res[p1 - p] = 0;

    return res;
}

static struct media *media_add(const char *p)
{
    const char *p1, *p2;
    struct media *m;
    int type;

    if (memcmp(p, "video", 5) == 0) {
        type = MEDIA_TYPE_VIDEO;
	p1 = p + 6;
    } else if (memcmp(p, "audio", 5) == 0) {
        type = MEDIA_TYPE_AUDIO;
	p1 = p + 6;
    } else if (memcmp(p, "text", 4) == 0) {
        type = MEDIA_TYPE_TEXT;
	p1 = p + 5;
    } else {
        return NULL;
    }

    m = malloc(sizeof(struct media));
    if (m == NULL) {
        return NULL;
    }
    memset(m, 0, sizeof(struct media));

    m->type = type;
    p2 = p1;
    while ((*p2 != 0) && (*p2 != 'R')) {
        p2++;
    }

    if (p2 != p1) {
        m->port = atoi(p1);
    } else {
        m->port = 0;
    }
    
    while ((*p2 != 0) && (*p2 != ' ')) {
        p2++;
    }

    if ((*p2 == 0) || (p2[1] == 0)) {
        free(m);

	return NULL;
    }

    m->payload_type = atoi(p2 + 1);

    return m;
}

static struct attribute *attribute_add(const char *p)
{
    const char *p1;
    struct attribute *attr;

    p1 = p;
    while ((*p1 != 0) && (*p1 != ':')) {
        p1++;
    }
    if (*p1 == 0) {
        return NULL;
    }

    attr = malloc(sizeof(struct attribute));
    if (attr == NULL) {
        return NULL;
    }

    attr->key = malloc(p1 - p + 1);
    if (attr->key == NULL) {
        free(attr);

	return NULL;
    }
    memcpy(attr->key, p, p1 - p);
    attr->key[p1 - p] = 0;
    
    p = line_skip(p1);
    p -= 1;
    p1 += 1;
    attr->val = malloc(p - p1 + 1);
    if (attr->val == NULL) {
        free(attr->key);
        free(attr);

	return NULL;
    }
    memcpy(attr->val, p1, p - p1);
    attr->val[p - p1] = 0;

    return attr;
}

static int addr_get(const char *buff, char *dst, int *ttl)
{
    while(*buff && (*buff != '/') && (*buff != '\n')) {
        *dst++ = *buff++;
    }
    *dst = 0;

    if (*buff == '/') {
        *ttl = atoi(buff + 1);
    } else {
        *ttl = 0;
    }

    return 0;
}

static char *sdp_print(char *buff, int size, const char *fmt, ...)
{
    va_list vl;
    int res;

    if (buff == NULL) {
        return NULL;
    }

    va_start(vl, fmt);
    res = vsnprintf(buff, size, fmt, vl);
    va_end(vl);
    if (res <= 0) {
        return NULL;
    }

    return buff + res;
}


static char *attribute_write(char *buff, int size, struct attribute *attr)
{
    return sdp_print(buff, size, "a=%s:%s\r\n", attr->key, attr->val);
}   

static char *media_write(char *buff, int size, struct media *m)
{
    const char *type;
    char *p;
    int i;

    switch(m->type) {
        case MEDIA_TYPE_VIDEO:
            type = "video";
            break;
        case MEDIA_TYPE_AUDIO:
            type = "audio";
            break;
        case MEDIA_TYPE_TEXT:
            type = "text";
            break;
        default:
            return NULL;
    }

    p = buff;
    p = sdp_print(buff, size, "m=%s %d RTP/AVP %d\r\n",
		               type, /*m->port*/ 0, m->payload_type);
    i = 0;
    while (m->attribute[i]) {
        p = attribute_write(p, size - (p - buff), m->attribute[i++]);
    }

    return p;
}

int sdp_get(char *buff, int size, struct sdp_descr *s)
{
    char *p;
    int i, video_streams;

    p = buff;
    p = sdp_print(p, size, "v=%d\r\n", s->sdp_version);
    p = sdp_print(p, size - (p - buff), "o=%s %d %d IN IPV4 %s\r\n",
                  s->user, s->rtp_id, s->version, s->src_addr);
    p = sdp_print(p, size - (p - buff), "t=%d %d\r\n",
                  s->start_time, s->end_time);
    p = sdp_print(p, size - (p - buff), "s=%s\r\n", s->name);

    i = 0;
    while (s->attribute[i]) {
        p = attribute_write(p, size - (p - buff), s->attribute[i++]);
    }

    /* If there are multiple video streams, only the first one is described
     * in the SDP (the other video streams are for different quality levels
     */
    i = 0;
    video_streams = 0;
    while (s->media[i]) {
	if ((video_streams == 0) || (s->media[i]->type == MEDIA_TYPE_AUDIO)) {
            p = media_write(p, size - (p - buff), s->media[i]);
            p = sdp_print(p, size - (p - buff), "a=control:TrackID=%d\n", i + 1);
	}
        if (s->media[i]->type == MEDIA_TYPE_VIDEO) {
            video_streams++;
	}
	i++;
    }

    if (p != NULL) {
        return 0;
    }

    return -1;
}

int sdp_set(struct sdp_descr *s, const char *sdp)
{
    const char *p;
    struct attribute **current_attr;
    char *current_dst;
    int *bitrate;
    int n_attr, n_media;
    int res, ttl;

    if (sdp == NULL) {
        fprintf(stderr, "Setting NULL SDP?\n");

        return -1;
    }
    n_attr = n_media = 0;
    p =  sdp;
    current_attr = s->attribute;
    current_dst = s->dst_addr;
    bitrate = NULL;
    while (*p) {
        if ((strlen(p) < 3) || (p[1] != '=')) {

            return -1;	/* FIXME: Leak? */
        }
        switch (p[0]) {
            case 'v':
                s->sdp_version = atoi(p + 2);
                break;
            case 'o':
                res = sscanf(p + 2, "%s %d %d IN IP4 %s",
                             s->user, &s->rtp_id, &s->version, s->src_addr);
                break;
            case 's':
                s->name = line_copy(p + 2);
                break;
            case 'i':
                s->info = line_copy(p + 2);
                break;
            case 't':
                res = sscanf(p + 2, "%d %d", &s->start_time, &s->end_time);
                break;
            case 'a':
                if (memcmp(p + 2, "control", 7)) {
                    current_attr[n_attr++] = attribute_add(p + 2);
                }
                break;
            case 'm':
		s->media[n_media] = media_add(p + 2);
		if (s->media[n_media]) {
                    current_attr = s->media[n_media]->attribute;
		    current_dst = s->media[n_media]->dst_addr;
                    bitrate = &s->media[n_media]->bitrate;
                    n_attr = 0;
                    n_media++;
                }
                break;
            case 'b':
		if (bitrate) {
                    *bitrate = atoi(p + 5);
		}
                break;
            case 'c':
                res = addr_get(p + 9, current_dst, &ttl);
                break;
        }
        p = line_skip(p);
        if (p == NULL) {
            return -1;	/* FIXME: Leak? */
        }
    }

    return 0;
}

void sdp_free(struct sdp_descr *s)
{
    int i, j;

    free(s->name);
    free(s->info);
    for (i = 0; i < MAX_MEDIA; i++) {
        if (s->media[i]) {
            j = 0;
            while (s->media[i]->attribute[j]) {
                free(s->media[i]->attribute[j]->key);
                free(s->media[i]->attribute[j]->val);
                free(s->media[i]->attribute[j++]);
            }
        }
        free(s->media[i]);
        s->media[i] = NULL;
    }
    j = 0;
    while (s->attribute[j]) {
        free(s->attribute[j]->key);
        free(s->attribute[j]->val);
        free(s->attribute[j++]);
    }
    /* FIXME: To be implemented! */
}
